'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var node_stream = require('node:stream');

class OctetStreamParser extends node_stream.PassThrough {
  constructor(options = {}) {
    super();
    this.globalOptions = { ...options };
  }
}

exports.default = OctetStreamParser;
